// Copyright (c) 2018 Kerlink SA. All rights reserved.

#include "filter.h"

#include "third-party/variant.h"

#include "utils/logger.h"
#include "utils/lorawan.h"
#include "utils/type_traits.h"

Filter::Filter(const Config &_config) : m_config(_config) {}

namespace {
template <typename T>
bool filterUplinkCRC(const Config &_config, const T &_data)
{
	switch (_data.crc) {
	case message::uplink::CRCStatus::CRC_OK:
		if (_config.filterCRCValid) {
			break;
		}

		logr::info << "Dropping message with valid CRC" << logr::endl;
		return false;

	case message::uplink::CRCStatus::CRC_ERROR:
		if (_config.filterCRCInvalid) {
			break;
		}

		logr::info << "Dropping message with invalid CRC" << logr::endl;
		return false;

	case message::uplink::CRCStatus::CRC_NONE:
		if (_config.filterCRCNone) {
			break;
		}

		logr::info << "Dropping message without CRC" << logr::endl;
		return false;
	}

	return true;
}

template <typename T>
bool filterUplinkLoRaWAN(const Config &_config, const T &_data)
{
	utils::LoRaWAN lorawan(_data.payload);

	if (!_config.filterLoRaWANDownlink &&
	    lorawan.direction == utils::LoRaWAN::DIR_DOWNLINK) {
		logr::info << "Dropping LoRaWAN downlink message" << logr::endl;
		return false;
	}

	return true;
}

struct UplinkVisitor
{
	const Config &m_config;

	UplinkVisitor(const Config &_config) : m_config(_config) {}

	bool operator()(...) { return true; }

	template <typename T,
		  typename = utils::enable_if_t<
			  utils::is_one_of<T,
					   message::lora::Uplink,
					   message::gfsk::Uplink>::value>>
	bool operator()(const T &_message)
	{
		if (_message.has_data) {
			auto data = _message.data;

			if (!filterUplinkCRC(m_config, data)) {
				return false;
			}

			if (data.crc != message::uplink::CRCStatus::CRC_ERROR) {
				if (!filterUplinkLoRaWAN(m_config, data)) {
					return false;
				}
			}
		}

		return true;
	}
};
}

bool Filter::input(const uplink_type &_message)
{
	return mpark::visit(UplinkVisitor(m_config), _message.messages);
}
